#!/usr/bin/env python3
import os

from flask import Flask, redirect, send_file
from flask.views import MethodView
import marshmallow as ma
from flask_smorest import Api, Blueprint, abort
from webargs.flaskparser import FlaskParser

from jass.config import config
from jass.models.phenotype import get_available_phenotypes
from jass.models.project import Project
from jass.tasks import create_project


class PhenotypeSchema(ma.Schema):
    id = ma.fields.String()
    consortium = ma.fields.String()
    outcome = ma.fields.String()
    full_name = ma.fields.String()
    type = ma.fields.String()
    ref = ma.fields.String()
    ref_link = ma.fields.String()
    data_link = ma.fields.String()
    data_path = ma.fields.String()


class ProjectParamsSchema(ma.Schema):
    class Meta:
        unknown = ma.EXCLUDE

    phenotypeID = ma.fields.List(ma.fields.String())


class LocalProjectParamsSchema(ma.Schema):
    class Meta:
        unknown = ma.EXCLUDE

    phenotypeID = ma.fields.List(ma.fields.String())
    chromosome = ma.fields.Integer()
    start = ma.fields.Integer()
    end = ma.fields.Integer()


class ProjectStatusSchema(ma.Schema):
    STATUS_VALUES = ["DOES_NOT_EXIST", "CREATING", "READY", "ERROR"]
    global_manhattan = ma.fields.String(validate=ma.validate.OneOf(STATUS_VALUES))
    quadrant_plot_status = ma.fields.String(validate=ma.validate.OneOf(STATUS_VALUES))
    worktable = ma.fields.String(validate=ma.validate.OneOf(STATUS_VALUES))


class ProjectSchema(ma.Schema):
    id = ma.fields.String()
    status = ma.fields.Nested(ProjectStatusSchema)
    phenotypes = ma.fields.List(ma.fields.Nested(PhenotypeSchema()))
    progress = ma.fields.String()


blp_phenotypes = Blueprint(
    "phenotypes",
    "phenotypes",
    url_prefix="/phenotypes",
    description="Operations on phenotypes",
)

blp_projects = Blueprint(
    "projects", "projects", url_prefix="/projects", description="Operations on projects"
)

blp_local_projects = Blueprint(
    "local_projects",
    "local_projects",
    url_prefix="/local_projects",
    description="Operations on local projects",
)


def get_phenotypes():
    return get_available_phenotypes(os.path.join(config["DATA_DIR"], "initTable.hdf5"))


@blp_phenotypes.route("")
class PhenotypesMethodView(MethodView):
    @blp_phenotypes.response(200, PhenotypeSchema(many=True))
    def get(self):
        """List phenotypes"""
        return get_phenotypes()


@blp_projects.route("")
class ProjectCreateMethodView(MethodView):
    @blp_projects.arguments(ProjectParamsSchema(), location="form")
    @blp_projects.response(200, ProjectSchema())
    def post(self, parameters):
        """List projects"""
        phenotype_ids = [
            phenotype_id
            for ids_with_commas in parameters["phenotypeID"]
            for phenotype_id in ids_with_commas.split(",")
        ]
        phenotypes = list(filter(lambda d: d.id in phenotype_ids, get_phenotypes()))
        return create_project([p.id for p in phenotypes], get_phenotypes())


@blp_local_projects.route("")
class LocalProjectCreateMethodView(MethodView):
    @blp_projects.arguments(LocalProjectParamsSchema(), location="form")
    @blp_projects.response(200, ProjectSchema())
    def post(self, parameters):
        """List projects"""
        phenotype_ids = [
            phenotype_id
            for ids_with_commas in parameters["phenotypeID"]
            for phenotype_id in ids_with_commas.split(",")
        ]
        phenotypes = list(filter(lambda d: d.id in phenotype_ids, get_phenotypes()))
        return create_project(
            [p.id for p in phenotypes],
            get_phenotypes(),
            str(parameters["chromosome"]),
            str(parameters["start"]),
            str(parameters["end"]),
        )


@blp_projects.route("/<project_id>")
class ProjectDetailMethodView(MethodView):
    @blp_projects.response(200, ProjectSchema())
    def get(self, project_id):
        return Project(id=project_id)


@blp_projects.route("/<project_id>/csv_status")
class ProjectCSVStatusMethodView(MethodView):
    def get(self, project_id):
        return Project(id=project_id).get_csv_file_generation()


@blp_projects.route("/<project_id>/summary")
class ProjectSummaryMethodView(MethodView):
    @blp_projects.response(200, ProjectSchema())
    def get(self, project_id):
        return Project(id=project_id).get_project_summary_statistics()


@blp_projects.route("/<project_id>/genome")
class ProjectGenomeMethodView(MethodView):

    # @blp_projects.response(200, headers={"Content-Type": "text/csv"})
    def get(self, project_id):
        try:
            return (
                Project(id=project_id).get_project_genomedata(),
                200,
                {"Content-Type": "text/plain; charset=utf-8"},
            )
        except FileNotFoundError:
            status = Project(id=project_id).status
            if status == Project.DOES_NOT_EXIST:
                return (
                    f"project {project_id} does not exist",
                    404,
                    {"Content-Type": "text/plain; charset=utf-8"},
                )
                abort(404)
            elif status["worktable"] == Project.CREATING:
                return (
                    "data not ready yet",
                    202,
                    {"Content-Type": "text/plain; charset=utf-8"},
                )
            else:
                abort(500)


@blp_projects.route("/<project_id>/genome_full")
class ProjectGenomeFullMethodView(MethodView):
    def get(self, project_id):
        Type_of_Analysis = Project(id=project_id).get_type_of_analysis()
        if(Type_of_Analysis == Project.LOCAL_ANALYSIS):
            Fichier = "local_analysis_result.csv"
        else:
            Fichier = "genome_full.csv"
        
        try:
            return send_file(
                Project(id=project_id).get_csv_path(),
                mimetype="text/csv",
                as_attachment=True,
                attachment_filename=Fichier,
            )
        except FileNotFoundError:
            status = Project(id=project_id).status
            if status == Project.DOES_NOT_EXIST:
                return (
                    f"project {project_id} does not exist",
                    404,
                    {"Content-Type": "text/plain; charset=utf-8"},
                )
            elif status["worktable"] == Project.CREATING:
                return (
                    "data not ready yet",
                    202,
                    {"Content-Type": "text/plain; charset=utf-8"},
                )
            else:
                abort(500)


@blp_projects.route("/<project_id>/globalmanhattan")
class ProjectGlobalManhattanMethodView(MethodView):
    def get(self, project_id):
        try:
            return send_file(
                Project(id=project_id).get_global_manhattan_plot_path(),
                mimetype="image/png",
            )
        except FileNotFoundError:
            status = Project(id=project_id).status
            if status == Project.DOES_NOT_EXIST:
                return (
                    f"project {project_id} does not exist",
                    404,
                    {"Content-Type": "text/plain; charset=utf-8"},
                )
            elif status["global_manhattan"] == Project.CREATING:
                return (
                    "data not ready yet",
                    202,
                    {"Content-Type": "text/plain; charset=utf-8"},
                )
            else:
                abort(500)


@blp_projects.route("/<project_id>/quadrant")
class ProjectQuadrantMethodView(MethodView):
    def get(self, project_id):
        try:
            return send_file(
                Project(id=project_id).get_quadrant_plot_path(),
                mimetype="image/png",
            )
        except FileNotFoundError:
            status = Project(id=project_id).status
            if status == Project.DOES_NOT_EXIST:
                return (
                    f"project {project_id} does not exist",
                    404,
                    {"Content-Type": "text/plain; charset=utf-8"},
                )
            elif status["quadrant_plot_status"] == Project.CREATING:
                return (
                    "data not ready yet",
                    202,
                    {"Content-Type": "text/plain; charset=utf-8"},
                )
            else:
                abort(500)


@blp_projects.route("/<project_id>/manhattan/<chromosome>/<region>")
class ProjectLocalManhattanMethodView(MethodView):
    def get(self, project_id, chromosome, region):
        try:
            return (
                Project(id=project_id).get_project_local_manhattan_data(
                    chromosome, region
                ),
                200,
                {"Content-Type": "text/plain; charset=utf-8"},
            )
        except FileNotFoundError:
            status = Project(id=project_id).status
            if status == Project.DOES_NOT_EXIST:
                return (
                    f"project {project_id} does not exist",
                    404,
                    {"Content-Type": "text/plain; charset=utf-8"},
                )
            elif status["worktable"] == Project.CREATING:
                return (
                    "data not ready yet",
                    202,
                    {"Content-Type": "text/plain; charset=utf-8"},
                )
            else:
                abort(500)


@blp_projects.route("/<project_id>/heatmap/<chromosome>/<region>")
class ProjectLocalHeatmapMethodView(MethodView):
    def get(self, project_id, chromosome, region):
        try:
            return (
                Project(id=project_id).get_project_local_heatmap_data(
                    chromosome, region
                ),
                200,
                {"Content-Type": "text/plain; charset=utf-8"},
            )
        except FileNotFoundError:
            status = Project(id=project_id).status
            if status == Project.DOES_NOT_EXIST:
                return (
                    f"project {project_id} does not exist",
                    404,
                    {"Content-Type": "text/plain; charset=utf-8"},
                )
            elif status["worktable"] == Project.CREATING:
                return (
                    "data not ready yet",
                    202,
                    {"Content-Type": "text/plain; charset=utf-8"},
                )
            else:
                abort(500)


@blp_projects.route("/<project_id>/zoom_manhattan")
class ProjectZoomManhattanMethodView(MethodView):
    def get(self, project_id):
        try:
            return (
                Project(id=project_id).get_project_local_manhattan_data(),
                200,
                {"Content-Type": "text/plain; charset=utf-8"},
            )
        except FileNotFoundError:
            status = Project(id=project_id).status
            if status == Project.DOES_NOT_EXIST:
                return (
                    f"project {project_id} does not exist",
                    404,
                    {"Content-Type": "text/plain; charset=utf-8"},
                )
            elif status["worktable"] == Project.CREATING:
                return (
                    "data not ready yet",
                    202,
                    {"Content-Type": "text/plain; charset=utf-8"},
                )
            else:
                abort(500)


@blp_projects.route("/<project_id>/zoom_heatmap")
class ProjectZoomHeatmapMethodView(MethodView):
    def get(self, project_id):
        try:
            return (
                Project(id=project_id).get_project_local_heatmap_data(),
                200,
                {"Content-Type": "text/plain; charset=utf-8"},
            )
        except FileNotFoundError:
            status = Project(id=project_id).status
            if status == Project.DOES_NOT_EXIST:
                return (
                    f"project {project_id} does not exist",
                    404,
                    {"Content-Type": "text/plain; charset=utf-8"},
                )
            elif status["worktable"] == Project.CREATING:
                return (
                    "data not ready yet",
                    202,
                    {"Content-Type": "text/plain; charset=utf-8"},
                )
            else:
                abort(500)


@blp_projects.route("/<project_id>/zoomplot")
class ProjectZoomPlotMethodView(MethodView):
    def get(self, project_id):
        try:
            return send_file(
                Project(id=project_id).get_zoom_plot_path(), mimetype="image/png"
            )
        except FileNotFoundError:
            status = Project(id=project_id).status
            if status == Project.DOES_NOT_EXIST:
                return (
                    f"project {project_id} does not exist",
                    404,
                    {"Content-Type": "text/plain; charset=utf-8"},
                )
            elif status["worktable"] == Project.CREATING:
                return (
                    "data not ready yet",
                    202,
                    {"Content-Type": "text/plain; charset=utf-8"},
                )
            else:
                abort(500)


class JassApp(Flask):
    """
    JassApp builds the JASS Flask application
    """

    def __init__(self):
        self.flask_app = Flask(__name__, static_url_path="", static_folder="static")
        self.flask_app.config["API_TITLE"] = "JASS API"
        self.flask_app.config["API_VERSION"] = "v2.0"
        self.flask_app.config["OPENAPI_VERSION"] = "3.0.2"
        self.flask_app.route("/")(self.redirect_to_index)
        self.api = Api(self.flask_app)

    def create_app(self):
        return self.flask_app

    def redirect_to_index(self):
        return redirect("index.html")

    def register_api_blueprint(self, blp):
        self.api.register_blueprint(blp, url_prefix=f"/api/{blp.url_prefix}")


jass_app = JassApp()
jass_app.register_api_blueprint(blp_phenotypes)
jass_app.register_api_blueprint(blp_projects)
jass_app.register_api_blueprint(blp_local_projects)
