import os
import sys
import traceback
from typing import List, Dict

from celery import Celery, group, chain
from flask import Flask

import jass.models.project
from jass.models.project import Project, get_file_building_tb_path

from jass.models.plots import (
    create_global_plot,
    create_local_plot,
    create_quadrant_plot,
    create_qq_plot,
)

from jass.models.worktable import create_worktable_file, create_genome_full_csv

from jass.models.phenotype import Phenotype
from jass.config import config


def make_celery(app):
    celery = Celery()
    if "CELERY_CONFIG_MODULE" in os.environ:
        celery.config_from_envvar("CELERY_CONFIG_MODULE")
    else:
        celery.config_from_object("jass.celeryconfig")
    celery.conf.update(app.config)
    TaskBase = celery.Task

    class ContextTask(TaskBase):
        abstract = True

        def __call__(self, *args, **kwargs):
            with app.app_context():
                return TaskBase.__call__(self, *args, **kwargs)

    celery.Task = ContextTask
    return celery


flask_app = Flask(__name__)
celery = make_celery(flask_app)


@celery.task
def create_project_worktable_file(
    phenotype_ids,
    init_file_path,
    project_hdf_path,
    remove_nan,
    stat,
    optim_na,
    csv_file,
    chunk_size,
    significance_treshold,
    post_filtering,
    delayed_gen_csv_file,
    chromosome,
    pos_Start,
    pos_End,
    custom_loadings,
):
    try:
        return create_worktable_file(
            phenotype_ids=phenotype_ids,
            init_file_path=init_file_path,
            project_hdf_path=project_hdf_path,
            remove_nan=remove_nan,
            stat=stat,
            optim_na=True,
            csv_file=csv_file,
            chunk_size=chunk_size,
            significance_treshold=significance_treshold,
            post_filtering=post_filtering,
            delayed_gen_csv_file=delayed_gen_csv_file,
            chromosome=chromosome,
            pos_Start=pos_Start,
            pos_End=pos_End,
            custom_loadings=custom_loadings,
        )
    except Exception as e:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        log_path = get_file_building_tb_path(project_hdf_path)
        log_fh = open(log_path, "w")
        traceback.print_exception(exc_type, exc_value, exc_traceback, file=log_fh)
        log_fh.close()


@celery.task
def create_project_global_plot(worktable_path, plot_path):
    try:
        create_global_plot(worktable_path, plot_path)
    except Exception as e:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        log_path = get_file_building_tb_path(plot_path)
        log_fh = open(log_path, "w")
        traceback.print_exception(exc_type, exc_value, exc_traceback, file=log_fh)
        log_fh.close()


@celery.task
def create_project_quadrant_plot(worktable_path, plot_path):
    try:
        create_quadrant_plot(worktable_path, plot_path)
    except Exception as e:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        log_path = get_file_building_tb_path(plot_path)
        log_fh = open(log_path, "w")
        traceback.print_exception(exc_type, exc_value, exc_traceback, file=log_fh)
        log_fh.close()


@celery.task
def create_project_zoom_plot(worktable_path, plot_path):
    try:
        create_local_plot(worktable_path, plot_path)
    except Exception as e:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        log_path = get_file_building_tb_path(plot_path)
        log_fh = open(log_path, "w")
        traceback.print_exception(exc_type, exc_value, exc_traceback, file=log_fh)
        log_fh.close()


@celery.task
def create_project_qq_plot(worktable_path, plot_path):
    try:
        create_qq_plot(worktable_path, plot_path)
    except Exception as e:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        log_path = get_file_building_tb_path(plot_path)
        log_fh = open(log_path, "w")
        traceback.print_exception(exc_type, exc_value, exc_traceback, file=log_fh)
        log_fh.close()


@celery.task
def create_project_csv_file(Nchunk, worktable_path, csv_file_path):
    try:
        create_genome_full_csv(worktable_path, csv_file_path, Nchunk=Nchunk)
    except Exception as e:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        log_path = get_file_building_tb_path(csv_file_path)
        log_fh = open(log_path, "w")
        traceback.print_exception(exc_type, exc_value, exc_traceback, file=log_fh)
        log_fh.close()


def launch_create_project(
    phenotype_ids,
    init_table_path,
    worktable_path,
    remove_nan=False,
    stat="jass.models.stats:omnibus_stat",
    csv_file=None,
    chunk_size=50,
    significance_treshold=5 * 10 ** -8,
    post_filtering=True,
    delayed_gen_csv_file=False,
    chromosome=None,
    start=None,
    end=None,
    custom_loadings=None,
    global_plot_path=None,
    quadrant_plot_path=None,
    zoom_plot_path=None,
    qq_plot_path=None,
):
    post_worktable_jobs = []
    if global_plot_path is not None:
        post_worktable_jobs.append(
            create_project_global_plot.si(worktable_path, global_plot_path)
        )
    if quadrant_plot_path is not None:
        post_worktable_jobs.append(
            create_project_quadrant_plot.si(worktable_path, quadrant_plot_path),
        )
    if zoom_plot_path is not None:
        post_worktable_jobs.append(
            create_project_zoom_plot.si(worktable_path, zoom_plot_path)
        )
    if qq_plot_path is not None:
        post_worktable_jobs.append(
            create_project_qq_plot.si(worktable_path, qq_plot_path)
        )
    if delayed_gen_csv_file and (csv_file is not None):
        post_worktable_jobs.append(create_project_csv_file.s(worktable_path, csv_file))
    post_worktable_tasks_group = group(post_worktable_jobs)
    main_wf = chain(
        create_project_worktable_file.si(
            phenotype_ids=phenotype_ids,
            init_file_path=init_table_path,
            project_hdf_path=worktable_path,
            remove_nan=remove_nan,
            stat=stat,
            optim_na=True,
            csv_file=csv_file,
            chunk_size=chunk_size,
            significance_treshold=significance_treshold,
            post_filtering=post_filtering,
            delayed_gen_csv_file=delayed_gen_csv_file,
            chromosome=chromosome,
            pos_Start=start,
            pos_End=end,
            custom_loadings=custom_loadings,
        ),
        post_worktable_tasks_group,
    )
    main_wf.delay()


def create_project(
    phenotype_ids: List[str],
    available_phenotypes: List[Phenotype],
    chromosome: str = None,
    start: str = None,
    end: str = None,
):

    available_phenotype_ids = [phenotype.id for phenotype in available_phenotypes]
    unavailable_requested_ids = set(phenotype_ids).difference(
        set(available_phenotype_ids)
    )
    if len(unavailable_requested_ids) > 0:
        raise Exception()  # FIXME with a nice exception
    phenotypes = [
        phenotype for phenotype in available_phenotypes if phenotype.id in phenotype_ids
    ]
    project = Project(
        phenotypes=phenotypes, chromosome=chromosome, start=start, end=end
    )

    if project.get_type_of_analysis() == Project.LOCAL_ANALYSIS:
        # Local Analysis
        global_plot_path = None
        quadrant_plot_path = None
        zoom_plot_path = project.get_zoom_plot_path()
        delayed_gen_csv_file = False
    else:
        # genome wide Analysis
        global_plot_path = project.get_global_manhattan_plot_path()
        quadrant_plot_path = project.get_quadrant_plot_path()
        zoom_plot_path = None
        delayed_gen_csv_file = True

    # if project does not exist
    if project.status == Project.DOES_NOT_EXIST:
        os.makedirs(project.get_folder_path())
        launch_create_project(
            phenotype_ids=phenotype_ids,
            init_table_path=os.path.join(config["DATA_DIR"], "initTable.hdf5"),
            worktable_path=project.get_worktable_path(),
            csv_file=project.get_csv_path(),
            global_plot_path=global_plot_path,
            quadrant_plot_path=quadrant_plot_path,
            zoom_plot_path=zoom_plot_path,
            delayed_gen_csv_file=delayed_gen_csv_file,
            chromosome=chromosome,
            start=start,
            end=end,
        )
    return project
